//////////////////////////////// Includes /////////////////////////////////////

#include "stdafx.h"
#include "WinUSBWrappers.h"
#include <initguid.h>
#include <usbiodef.h>
#include <ctime>
#include <array>


//////////////////////////////// Macros / Defines /////////////////////////////

//Defines / Structs as taken from https://store.osr.com/wp-content/uploads/sites/2/2014/08/OSRFX2_35.pdf, 
//https://www.osronline.com/article.cfm?article=532 and 
//http://msdn.microsoft.com/en-us/library/windows/hardware/ff540174(v=vs.85).aspx
#define FX2_READ_7SEGMENT_DISPLAY 0xD4
#define FX2_SET_BARGRAPH_DISPLAY  0xD8
#define FX2_READ_SWITCHES         0xD6
#define FX2_READ_BARGRAPH_DISPLAY 0xD7
#define FX2_SET_7SEGMENT_DISPLAY  0xDB
#define CYPRESS_SEMICONDUCTOR_VID 0x0547
#define FX2_PID                   0x1002
#define FX2_INTERRUPT_ENDPOINT    0x81
#define FX2_BULK_IN_ENDPOINT      0x88
#define FX2_BULK_OUT_ENDPOINT     0x06

struct USB_DEVICE_REQUESTTYPE
{
  union
  {
    BYTE bmRequestType;
    struct
    {
      BYTE Recipient     : 5;
      BYTE Type          : 2;
      BYTE DataDirection : 1;
    } Bits;
  };
};

struct FX2_BAR_GRAPH_STATE
{
  union
  {
    UCHAR StateAsUChar;
    struct
    {
      UCHAR Bar4: 1;
      UCHAR Bar5: 1;
      UCHAR Bar6: 1;
      UCHAR Bar7: 1;
      UCHAR Bar8: 1;
      UCHAR Bar1: 1;
      UCHAR Bar2: 1;
      UCHAR Bar3: 1;
    } Bits;
  };
};

struct FX2_SWITCH_STATE
{
  union
  {
    UCHAR SwitchesAsUChar;
    struct 
    {
      UCHAR Switch8 : 1;
      UCHAR Switch7 : 1;
      UCHAR Switch6 : 1;
      UCHAR Switch5 : 1;
      UCHAR Switch4 : 1;
      UCHAR Switch3 : 1;
      UCHAR Switch2 : 1;
      UCHAR Switch1 : 1;
    } Bits;
  };
};

//Note this is the interface GUID you have used from your inf file when associating WinUSB with your FX2 device.
//You will of course need to change the value here to match the value in your inf file.
DEFINE_GUID(GUID_DEVINTERFACE_OSR2_FX2, 0x1A7CDD86, 0x3100, 0x467D, 0x82, 0x29, 0x10, 0x2B, 0x91, 0xF8, 0x8C, 0xA8);
#define GUID_DEVINTERFACE_OSR2_FX2_STR _T("{1A7CDD86-3100-467D-8229-102B91F88CA8}")


//////////////////////////////// Implementation ///////////////////////////////

void LightBarGraph(_Inout_ WinUSB::CDevice& device, _In_ UCHAR Light) noexcept
{
  USB_DEVICE_REQUESTTYPE requestType;
  requestType.Bits.DataDirection = 0; //Host to Device
  requestType.Bits.Recipient = 0; //Device
  requestType.Bits.Type = 2; //Vendor
  WINUSB_SETUP_PACKET packet;
  packet.RequestType = requestType.bmRequestType;
  packet.Request = FX2_SET_BARGRAPH_DISPLAY;
  packet.Index = 0;
  FX2_BAR_GRAPH_STATE bar;
  bar.StateAsUChar = Light;
  packet.Length = sizeof(bar);
  packet.Value = 0;
  ULONG nBytesReturned = 0;
#pragma warning(suppress: 26490)
  const BOOL bSuccess = device.ControlTransfer(packet, reinterpret_cast<PUCHAR>(&bar), sizeof(bar), &nBytesReturned, nullptr);
#pragma warning(suppress: 26477)
  ATLASSERT(bSuccess);
  UNREFERENCED_PARAMETER(bSuccess);
}

void LightSevenSegment(_Inout_ WinUSB::CDevice& device, _In_ UCHAR sevenSegment) noexcept
{
  USB_DEVICE_REQUESTTYPE requestType;
  requestType.Bits.DataDirection = 0; //Host to Device
  requestType.Bits.Recipient = 0; //Device
  requestType.Bits.Type = 2; //Vendor
  WINUSB_SETUP_PACKET packet;
  packet.RequestType = requestType.bmRequestType; 
  packet.Request = FX2_SET_7SEGMENT_DISPLAY;
  packet.Index = 0;
  packet.Length = sizeof(sevenSegment);
  packet.Value = 0;
  ULONG nBytesReturned = 0;
  const BOOL bSuccess = device.ControlTransfer(packet, &sevenSegment, sizeof(sevenSegment), &nBytesReturned, nullptr);
#pragma warning(suppress: 26477)
  ATLASSERT(bSuccess);
  UNREFERENCED_PARAMETER(bSuccess);
}

int _tmain()
{
#pragma warning(suppress: 26472)
  srand(static_cast<unsigned int>(time(nullptr)));

  //Enumerate all the USB devices names on the local machine and log the details for each device while we are at it
  _tprintf(_T("Enumerating USB devices using standard USB Device interface GUID on this computer\n"));
  WinUSB::StringArray deviceNames;
  BOOL bSuccess{WinUSB::CDevice::EnumerateDevices(&GUID_DEVINTERFACE_USB_DEVICE, deviceNames, DIGCF_PRESENT | DIGCF_DEVICEINTERFACE, nullptr, [](_In_ HDEVINFO hDeviceInfoSet, _In_ PSP_DEVICE_INTERFACE_DETAIL_DATA /*pDeviceInterfaceDetailData*/, _In_ PSP_DEVINFO_DATA pDeviceInfoData, _In_z_ LPCTSTR pszDevicePath, _In_ DWORD_PTR /*dwItemData*/) noexcept //NOLINT(clang-analyzer-deadcode.DeadStores)
  {
    _tprintf(_T(" %s\n"), pszDevicePath);
    std::array<TCHAR, 4096> buffer;
#pragma warning(suppress: 26446)
    buffer[0] = _T('\0');
#pragma warning(suppress: 26472 26490)
    if (SetupDiGetDeviceRegistryProperty(hDeviceInfoSet, pDeviceInfoData, SPDRP_FRIENDLYNAME, nullptr, reinterpret_cast<PBYTE>(buffer.data()), static_cast<DWORD>(buffer.size()), nullptr))
      _tprintf(_T(" Friendly name:%s\n"), buffer.data());
#pragma warning(suppress: 26472 26490)
    if (SetupDiGetDeviceRegistryProperty(hDeviceInfoSet, pDeviceInfoData, SPDRP_CLASS, nullptr, reinterpret_cast<PBYTE>(buffer.data()), static_cast<DWORD>(buffer.size()), nullptr))
      _tprintf(_T(" Class:%s\n"), buffer.data());
#pragma warning(suppress: 26472 26490)
    if (SetupDiGetDeviceRegistryProperty(hDeviceInfoSet, pDeviceInfoData, SPDRP_CLASSGUID, nullptr, reinterpret_cast<PBYTE>(buffer.data()), static_cast<DWORD>(buffer.size()), nullptr))
      _tprintf(_T(" Class GUID:%s\n"), buffer.data());
#pragma warning(suppress: 26472 26490)
    if (SetupDiGetDeviceRegistryProperty(hDeviceInfoSet, pDeviceInfoData, SPDRP_DEVICEDESC, nullptr, reinterpret_cast<PBYTE>(buffer.data()), static_cast<DWORD>(buffer.size()), nullptr))
      _tprintf(_T(" Device Description:%s\n"), buffer.data());
#pragma warning(suppress: 26472 26490)
    if (SetupDiGetDeviceRegistryProperty(hDeviceInfoSet, pDeviceInfoData, SPDRP_DRIVER, nullptr, reinterpret_cast<PBYTE>(buffer.data()), static_cast<DWORD>(buffer.size()), nullptr))
      _tprintf(_T(" Driver:%s\n"), buffer.data());
#pragma warning(suppress: 26472 26490)
    if (SetupDiGetDeviceRegistryProperty(hDeviceInfoSet, pDeviceInfoData, SPDRP_ENUMERATOR_NAME, nullptr, reinterpret_cast<PBYTE>(buffer.data()), static_cast<DWORD>(buffer.size()), nullptr))
      _tprintf(_T(" Enumerator Name:%s\n"), buffer.data());
#pragma warning(suppress: 26472 26490)
    if (SetupDiGetDeviceRegistryProperty(hDeviceInfoSet, pDeviceInfoData, SPDRP_LOCATION_INFORMATION, nullptr, reinterpret_cast<PBYTE>(buffer.data()), static_cast<DWORD>(buffer.size()), nullptr))
      _tprintf(_T(" Location Information:%s\n"), buffer.data());
#pragma warning(suppress: 26472 26490)
    if (SetupDiGetDeviceRegistryProperty(hDeviceInfoSet, pDeviceInfoData, SPDRP_MFG, nullptr, reinterpret_cast<PBYTE>(buffer.data()), static_cast<DWORD>(buffer.size()), nullptr))
      _tprintf(_T(" Manufacturer:%s\n"), buffer.data());
#pragma warning(suppress: 26472 26490)
    if (SetupDiGetDeviceRegistryProperty(hDeviceInfoSet, pDeviceInfoData, SPDRP_PHYSICAL_DEVICE_OBJECT_NAME, nullptr, reinterpret_cast<PBYTE>(buffer.data()), static_cast<DWORD>(buffer.size()), nullptr))
      _tprintf(_T(" PDO:%s\n"), buffer.data());
#pragma warning(suppress: 26472 26490)
    if (SetupDiGetDeviceRegistryProperty(hDeviceInfoSet, pDeviceInfoData, SPDRP_SECURITY_SDS, nullptr, reinterpret_cast<PBYTE>(buffer.data()), static_cast<DWORD>(buffer.size()), nullptr))
      _tprintf(_T(" Security Descriptor:%s\n"), buffer.data());
#pragma warning(suppress: 26472 26490)
    if (SetupDiGetDeviceRegistryProperty(hDeviceInfoSet, pDeviceInfoData, SPDRP_SERVICE, nullptr, reinterpret_cast<PBYTE>(buffer.data()), static_cast<DWORD>(buffer.size()), nullptr))
      _tprintf(_T(" Service:%s\n"), buffer.data());
    _tprintf(_T("\n"));

    return true;
  })};
#pragma warning(suppress: 26477)
  ATLASSERT(bSuccess);

  _tprintf(_T("Enumerating Cypress Semiconductor / Anchor Chips (the FX2 VID) devices on this computer\n"));
  bSuccess = WinUSB::CDevice::EnumerateDevices(&GUID_DEVINTERFACE_USB_DEVICE, CYPRESS_SEMICONDUCTOR_VID, deviceNames); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(bSuccess);
  for (const auto& deviceName : deviceNames)
    _tprintf(_T(" %s\n"), deviceName.c_str());
  if (deviceNames.size() == 0)
      _tprintf(_T(" No devices found\n"));

  _tprintf(_T("Enumerating OSR USX FX2 devices using specific interface GUID of %s on this computer. Note this will only work if your FX2 device has this interface GUID in its inf file\n"), GUID_DEVINTERFACE_OSR2_FX2_STR);
  bSuccess = WinUSB::CDevice::EnumerateDevices(&GUID_DEVINTERFACE_OSR2_FX2, deviceNames); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(bSuccess);
  for (const auto& deviceName : deviceNames)
    _tprintf(_T(" %s\n"), deviceName.c_str());
  if (deviceNames.size() == 0)
    _tprintf(_T(" No devices found\n"));

  _tprintf(_T("Enumerating OSR USX FX2 devices using standard USB Device interface GUID, and FX2 VID and PID on this computer\n"));
  bSuccess = WinUSB::CDevice::EnumerateDevices(&GUID_DEVINTERFACE_USB_DEVICE, CYPRESS_SEMICONDUCTOR_VID, FX2_PID, deviceNames); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(bSuccess);
  WinUSB::String sDevicePath;
  for (const auto& deviceName : deviceNames)
    _tprintf(_T(" %s\n"), deviceName.c_str());
  if (deviceNames.size() == 0)
  {
    _tprintf(_T(" No devices found\n"));
    return 1;
  }
  else
#pragma warning(suppress: 26446)
    sDevicePath = deviceNames[0];

  WinUSB::CDevice device;
  bSuccess = device.Initialize(_T("\\\\?\\usb#vid_1310&pid_0001#ae9a330d0b00#{a5dcbf10-6530-11d2-901f-00c04fb951ed}")); //NOLINT(clang-analyzer-deadcode.DeadStores)

  WinUSB::CDevice device2;
  _tprintf(_T("Attempting to open device \\\\?\\usb#vid_0547&pid_1002#7&3700aea3&0&3#{a5dcbf10-6530-11d2-901f-00c04fb951XX}, should fail!\n"));
  bSuccess = device2.Initialize(_T("\\\\?\\usb#vid_0547&pid_1002#7&3700aea3&0&3#{a5dcbf10-6530-11d2-901f-00c04fb951XX}")); //NOLINT(clang-analyzer-deadcode.DeadStores)
  _tprintf(_T("Attempting to open device %s, should succeed\n"), sDevicePath.c_str());
  bSuccess = device2.Initialize(sDevicePath.c_str()); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(bSuccess);

  //Let's query for the device descriptor
  _tprintf(_T("Getting device descriptor\n"));
  USB_DEVICE_DESCRIPTOR deviceDescriptor{};
  ULONG nTransferred{0};
#pragma warning(suppress: 26490)
  bSuccess = device2.GetDescriptor(USB_DEVICE_DESCRIPTOR_TYPE, 0, 0, reinterpret_cast<PUCHAR>(&deviceDescriptor), sizeof(deviceDescriptor), &nTransferred); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(bSuccess);
#pragma warning(suppress: 26472)
  _tprintf(_T(" bLength: 0x%02X\n"), static_cast<int>(deviceDescriptor.bLength));
#pragma warning(suppress: 26472)
  _tprintf(_T(" bDescriptorType: 0x%02X\n"), static_cast<int>(deviceDescriptor.bDescriptorType));
#pragma warning(suppress: 26472)
  _tprintf(_T(" bcdUSB: 0x%04X\n"), static_cast<int>(deviceDescriptor.bcdUSB));
#pragma warning(suppress: 26472)
  _tprintf(_T(" bDeviceClass: 0x%02X\n"), static_cast<int>(deviceDescriptor.bDeviceClass));
#pragma warning(suppress: 26472)
  _tprintf(_T(" bDeviceSubClass: 0x%02X\n"), static_cast<int>(deviceDescriptor.bDeviceSubClass));
#pragma warning(suppress: 26472)
  _tprintf(_T(" bDeviceProtocol: 0x%02X\n"), static_cast<int>(deviceDescriptor.bDeviceProtocol));
#pragma warning(suppress: 26472)
  _tprintf(_T(" bMaxPacketSize0: 0x%02X bytes\n"), static_cast<int>(deviceDescriptor.bMaxPacketSize0));
#pragma warning(suppress: 26472)
  _tprintf(_T(" idVendor: 0x%04X\n"), static_cast<int>(deviceDescriptor.idVendor));
#pragma warning(suppress: 26472)
  _tprintf(_T(" idProduct: 0x%04X\n"), static_cast<int>(deviceDescriptor.idProduct));
  if (deviceDescriptor.iProduct)
  {
    std::array<UCHAR, 1024> stringDescriptor{};
    nTransferred = 0;
#pragma warning(suppress: 26472)
    bSuccess = device2.GetDescriptor(USB_STRING_DESCRIPTOR_TYPE, deviceDescriptor.iProduct, 0, stringDescriptor.data(), static_cast<ULONG>(stringDescriptor.size()), &nTransferred); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
    ATLASSERT(bSuccess);
#pragma warning(suppress: 26490)
    auto pStringDescriptor{reinterpret_cast<USB_STRING_DESCRIPTOR*>(stringDescriptor.data())};
#pragma warning(suppress: 26446 26482)
    pStringDescriptor->bString[(pStringDescriptor->bLength - 2) /sizeof(WCHAR)] = L'\0';
#pragma warning(suppress: 26485)
    wprintf(L"  Product name: %s\n", pStringDescriptor->bString);
  }
  else
    _tprintf(_T("  No product name was provided\n"));
#pragma warning(suppress: 26472)
  _tprintf(_T(" bcdDevice: 0x%04X\n"), static_cast<int>(deviceDescriptor.bcdDevice));
#pragma warning(suppress: 26472)
  _tprintf(_T(" iManufacturer: 0x%02X\n"), static_cast<int>(deviceDescriptor.iManufacturer));
  if (deviceDescriptor.iManufacturer)
  {
    std::array<UCHAR, 1024> stringDescriptor{};
    nTransferred = 0;
#pragma warning(suppress: 26472)
    bSuccess = device2.GetDescriptor(USB_STRING_DESCRIPTOR_TYPE, deviceDescriptor.iManufacturer, 0, stringDescriptor.data(), static_cast<ULONG>(stringDescriptor.size()), &nTransferred); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
    ATLASSERT(bSuccess);
#pragma warning(suppress: 26490)
    auto pStringDescriptor{reinterpret_cast<USB_STRING_DESCRIPTOR*>(stringDescriptor.data())};
#pragma warning(suppress: 26446 26482)
    pStringDescriptor->bString[(pStringDescriptor->bLength - 2) /sizeof(WCHAR)] = L'\0';
#pragma warning(suppress: 26485)
    wprintf(L"  Manufacturer name: %s\n", pStringDescriptor->bString);
  }
  else
    _tprintf(_T("  No manufacturer name was provided\n"));
#pragma warning(suppress: 26472)
  _tprintf(_T(" iProduct: 0x%02X\n"), static_cast<int>(deviceDescriptor.iProduct));
#pragma warning(suppress: 26472)
  _tprintf(_T(" iSerialNumber: 0x%02X\n"), static_cast<int>(deviceDescriptor.iSerialNumber));
  if (deviceDescriptor.iSerialNumber)
  {
    std::array<UCHAR, 1024> stringDescriptor{};
    nTransferred = 0;
#pragma warning(suppress: 26472)
    bSuccess = device2.GetDescriptor(USB_STRING_DESCRIPTOR_TYPE, deviceDescriptor.iSerialNumber, 0, stringDescriptor.data(), static_cast<ULONG>(stringDescriptor.size()), &nTransferred); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
    ATLASSERT(bSuccess);
#pragma warning(suppress: 26490)
    auto pStringDescriptor{reinterpret_cast<USB_STRING_DESCRIPTOR*>(stringDescriptor.data())};
#pragma warning(suppress: 26446 26482)
    pStringDescriptor->bString[(pStringDescriptor->bLength - 2) /sizeof(WCHAR)] = L'\0';
#pragma warning(suppress: 26485)
    wprintf(L"  Serial Number: %s\n", pStringDescriptor->bString);
  }
  else
    _tprintf(_T("  No serial number was provided\n"));
#pragma warning(suppress: 26472)
  _tprintf(_T(" bNumConfigurations: %d\n"), static_cast<int>(deviceDescriptor.bNumConfigurations));

  //Let's query for the configuration descriptor
  _tprintf(_T("Getting configuration descriptor\n"));
  USB_CONFIGURATION_DESCRIPTOR configDescriptor{};
  ULONG nBytesReturned = 0;
#pragma warning(suppress: 26490)
  bSuccess = device2.GetDescriptor(USB_CONFIGURATION_DESCRIPTOR_TYPE, 0, 0, reinterpret_cast<PUCHAR>(&configDescriptor), sizeof(configDescriptor), &nTransferred); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(bSuccess);
#pragma warning(suppress: 26472)
  _tprintf(_T(" bLength: %d\n"), static_cast<int>(configDescriptor.bLength));
#pragma warning(suppress: 26472)
  _tprintf(_T(" bDescriptorType: 0x%02X\n"), static_cast<int>(configDescriptor.bDescriptorType));
#pragma warning(suppress: 26472)
  _tprintf(_T(" wTotalLength: 0x%04X\n"), static_cast<int>(configDescriptor.wTotalLength));
#pragma warning(suppress: 26472)
  _tprintf(_T(" bNumInterfaces: 0x%02X\n"), static_cast<int>(configDescriptor.bNumInterfaces));
#pragma warning(suppress: 26472)
  _tprintf(_T(" bConfigurationValue: 0x%02X\n"), static_cast<int>(configDescriptor.bConfigurationValue));
#pragma warning(suppress: 26472)
  _tprintf(_T(" iConfiguration: 0x%02X\n"), static_cast<int>(configDescriptor.iConfiguration));
  if (configDescriptor.iConfiguration)
  {
    std::array<UCHAR, 1024> stringDescriptor{};
    nTransferred = 0;
#pragma warning(suppress: 26472)
    bSuccess = device2.GetDescriptor(USB_STRING_DESCRIPTOR_TYPE, configDescriptor.iConfiguration, 0, stringDescriptor.data(), static_cast<ULONG>(stringDescriptor.size()), &nTransferred); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
    ATLASSERT(bSuccess);
#pragma warning(suppress: 26490)
    auto pStringDescriptor{reinterpret_cast<USB_STRING_DESCRIPTOR*>(stringDescriptor.data())};
#pragma warning(suppress: 26446 26482)
    pStringDescriptor->bString[(pStringDescriptor->bLength - 2) /sizeof(WCHAR)] = L'\0';
#pragma warning(suppress: 26485)
    wprintf(L"  Configuration name: %s\n", pStringDescriptor->bString);
  }
#pragma warning(suppress: 26472)
  _tprintf(_T(" bmAttributes: 0x%02X\n"), static_cast<int>(configDescriptor.bmAttributes));
#pragma warning(suppress: 26472)
  _tprintf(_T(" MaxPower: 0x%02X\n"), static_cast<int>(configDescriptor.MaxPower));

  //Let's query the device for info
  _tprintf(_T("Querying device speed\n"));
  UCHAR deviceSpeed{0};
  ULONG nBufferLen{sizeof(deviceSpeed)};
  bSuccess = device2.QueryDeviceInformation(DEVICE_SPEED, &nBufferLen, &deviceSpeed); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(bSuccess);
  switch (deviceSpeed)
  {
    case LowSpeed:
    {
      _tprintf(_T(" The device is operating at low speed\n"));
      break;
    }
    case FullSpeed:
    {
      _tprintf(_T(" The device is operating at full speed\n"));
      break;
    }
    case HighSpeed:
    {
      _tprintf(_T(" The device is operating at high speed\n"));
      break;
    }
    default:
    {
#pragma warning(suppress: 26472)
      _tprintf(_T(" The device is operating at unknown %d speed\n"), static_cast<int>(deviceSpeed));
      break;
    }
  }
  UCHAR alternateSetting{0};
  bSuccess = device2.GetCurrentAlternateSetting(&alternateSetting); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(bSuccess);
#pragma warning(suppress: 26472)
  _tprintf(_T("Current alternate setting %d\n"), static_cast<int>(alternateSetting));
  //bSuccess = device2.SetCurrentAlternateSetting(alternateSetting); //NOLINT(clang-analyzer-deadcode.DeadStores)
  //ATLASSERT(bSuccess);

  //Let's query for interface and pipe information
  _tprintf(_T("Querying device interfaces and pipes\n"));
  bSuccess = TRUE;
  for (UCHAR interfaceNumber=0; (interfaceNumber<=0xFF) && bSuccess; interfaceNumber++) 
  {
    USB_INTERFACE_DESCRIPTOR interfaceDescriptor{};
    bSuccess = device2.QueryInterfaceSettings(interfaceNumber, &interfaceDescriptor);
    if (!bSuccess)
    {
#pragma warning(suppress: 26477)
      ATLASSERT(GetLastError() == ERROR_NO_MORE_ITEMS);
    }
    else
    {
#pragma warning(suppress: 26472)
      _tprintf(_T(" Interface %d details\n"), static_cast<int>(interfaceNumber));
#pragma warning(suppress: 26472)
      _tprintf(_T("  bLength: %d\n"), static_cast<int>(interfaceDescriptor.bLength));
#pragma warning(suppress: 26472)
      _tprintf(_T("  bDescriptorType: 0x%02X\n"),  static_cast<int>(interfaceDescriptor.bDescriptorType));
#pragma warning(suppress: 26472)
      _tprintf(_T("  bInterfaceNumber: 0x%02X\n"), static_cast<int>(interfaceDescriptor.bInterfaceNumber));
#pragma warning(suppress: 26472)
      _tprintf(_T("  bAlternateSetting: 0x%02X\n"), static_cast<int>(interfaceDescriptor.bAlternateSetting));
#pragma warning(suppress: 26472)
      _tprintf(_T("  bNumEndpoints: 0x%02X\n"), static_cast<int>(interfaceDescriptor.bNumEndpoints));
#pragma warning(suppress: 26472)
      _tprintf(_T("  bInterfaceClass: 0x%02X\n"), static_cast<int>(interfaceDescriptor.bInterfaceClass));
#pragma warning(suppress: 26472)
      _tprintf(_T("  bInterfaceSubClass: 0x%02X\n"), static_cast<int>(interfaceDescriptor.bInterfaceSubClass));
#pragma warning(suppress: 26472)
      _tprintf(_T("  bInterfaceProtocol: 0x%02X\n"), static_cast<int>(interfaceDescriptor.bInterfaceProtocol));
#pragma warning(suppress: 26472)
      _tprintf(_T("  iInterface: 0x%02X\n"), static_cast<int>(interfaceDescriptor.iInterface));

      BOOL bContinueIteratingPipes{TRUE};
      for (UCHAR pipeNumber=0; (pipeNumber<=0xFF) && bContinueIteratingPipes; pipeNumber++) 
      {
        WINUSB_PIPE_INFORMATION pipeInformation{};
        bContinueIteratingPipes = device2.QueryPipe(interfaceNumber, pipeNumber, &pipeInformation);
        if (!bContinueIteratingPipes)
        {
#pragma warning(suppress: 26477)
          ATLASSERT(GetLastError() == ERROR_NO_MORE_ITEMS);
        }
        else
        {
          WINUSB_PIPE_INFORMATION_EX pipeInformationEx{};
          const BOOL bSuccess2{device2.QueryPipEx(interfaceNumber, pipeNumber, &pipeInformationEx)};
#pragma warning(suppress: 26477)
          ATLASSERT(bSuccess2);
          UNREFERENCED_PARAMETER(bSuccess2);

          switch (pipeInformation.PipeType)
          {
            case UsbdPipeTypeControl:
            {
#pragma warning(suppress: 26472)
              _tprintf(_T("  Control Pipe, ID:0x%02X, Max Packet Size:%d, Interval:%d, Max bytes per interval:%d\n"), static_cast<int>(pipeInformation.PipeId), static_cast<int>(pipeInformation.MaximumPacketSize), static_cast<int>(pipeInformation.Interval), static_cast<int>(pipeInformationEx.MaximumBytesPerInterval));
              break;
            }
            case UsbdPipeTypeIsochronous:
            {
#pragma warning(suppress: 26472)
              _tprintf(_T("  Isochronous Pipe, ID:0x%02X, Max Packet Size:%d, Interval:%d, Max bytes per interval:%d\n"), static_cast<int>(pipeInformation.PipeId), static_cast<int>(pipeInformation.MaximumPacketSize), static_cast<int>(pipeInformation.Interval), static_cast<int>(pipeInformationEx.MaximumBytesPerInterval));
              break;
            }
            case UsbdPipeTypeBulk:
            {
              if (USB_ENDPOINT_DIRECTION_IN(pipeInformation.PipeId))
#pragma warning(suppress: 26472)
                _tprintf(_T("  Bulk IN Pipe, ID:0x%02X, Max Packet Size:%d, Interval:%d, Max bytes per interval:%d\n"), static_cast<int>(pipeInformation.PipeId), static_cast<int>(pipeInformation.MaximumPacketSize), static_cast<int>(pipeInformation.Interval), static_cast<int>(pipeInformationEx.MaximumBytesPerInterval));
              if (USB_ENDPOINT_DIRECTION_OUT(pipeInformation.PipeId))
#pragma warning(suppress: 26472)
                _tprintf(_T("  Bulk OUT Pipe, ID:0x%02X, Max Packet Size:%d, Interval:%d, Max bytes per interval:%d\n"), static_cast<int>(pipeInformation.PipeId), static_cast<int>(pipeInformation.MaximumPacketSize), static_cast<int>(pipeInformation.Interval), static_cast<int>(pipeInformationEx.MaximumBytesPerInterval));
              break;
            }
            case UsbdPipeTypeInterrupt:
            {
#pragma warning(suppress: 26472)
              _tprintf(_T("  Interrupt Pipe, ID:0x%02X, Max Packet Size:%d, Interval:%d, Max bytes per interval:%d\n"), static_cast<int>(pipeInformation.PipeId), static_cast<int>(pipeInformation.MaximumPacketSize), static_cast<int>(pipeInformation.Interval), static_cast<int>(pipeInformationEx.MaximumBytesPerInterval));
              break;
            }
            default:
            {
#pragma warning(suppress: 26472)
              _tprintf(_T("  Unknown [%d] Pipe, ID:0x%02X, Max Packet Size:%d, Interval:%d, Max bytes per interval:%d\n"), static_cast<int>(pipeInformation.PipeType), static_cast<int>(pipeInformation.PipeId), static_cast<int>(pipeInformation.MaximumPacketSize), static_cast<int>(pipeInformation.Interval), static_cast<int>(pipeInformationEx.MaximumBytesPerInterval));
              break;
            }
          }
        }
      }
    }
  }

  //Test out various parts of the WinUSB::CDevice class
  WinUSB::CDevice device4;
  bSuccess = device2.GetAssociatedInterface(0, device4); //NOLINT(clang-analyzer-deadcode.DeadStores)
  WINUSB_INTERFACE_HANDLE hHandle{device2};
  UNREFERENCED_PARAMETER(hHandle);
  HANDLE hHandle2{device2.GetDeviceHandle()};
  UNREFERENCED_PARAMETER(hHandle2);
  ULONG nCurrentFrameNumber{0};
  LARGE_INTEGER timeStamp{};
  bSuccess = device2.GetCurrentFrameNumber(&nCurrentFrameNumber, &timeStamp); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(bSuccess);
  std::array<UCHAR, 1024> isoBuffer{};
  WINUSB_ISOCH_BUFFER_HANDLE BufferHandle{nullptr};
#pragma warning(suppress: 26472)
  bSuccess = device2.RegisterIsochBuffer(FX2_BULK_IN_ENDPOINT, isoBuffer.data(), static_cast<ULONG>(isoBuffer.size()), &BufferHandle); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(bSuccess);
  WinUsb_UnregisterIsochBuffer(BufferHandle);
  DWORD dwTimeout{0};
  DWORD dwSize{sizeof(dwTimeout)};
  bSuccess = device2.GetPipePolicy(FX2_BULK_IN_ENDPOINT, PIPE_TRANSFER_TIMEOUT, &dwSize, &dwTimeout); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(bSuccess);
  bSuccess = device2.SetPipePolicy(FX2_BULK_IN_ENDPOINT, PIPE_TRANSFER_TIMEOUT, dwSize, &dwTimeout); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(bSuccess);

  __if_exists(WinUSB::CDevice::StartTrackingForTimeSync)
  {
    USB_START_TRACKING_FOR_TIME_SYNC_INFORMATION startTrackingInfo{};
    startTrackingInfo.TimeTrackingHandle = INVALID_HANDLE_VALUE;
    startTrackingInfo.IsStartupDelayTolerable = TRUE;
    bSuccess = device2.StartTrackingForTimeSync(&startTrackingInfo);

    /*
    __if_exists(WinUSB::CDevice::GetCurrentFrameNumberAndQpc)
    {
      USB_FRAME_NUMBER_AND_QPC_FOR_TIME_SYNC_INFORMATION frameQpcInfo{};
      frameQpcInfo.TimeTrackingHandle = device2;
      const BOOL bSuccess2 = device2.GetCurrentFrameNumberAndQpc(&frameQpcInfo);
      UNREFERENCED_PARAMETER(bSuccess2);
    } //__if_exists(WinUSB::CDevice::GetCurrentFrameNumberAndQpc)
    */

    if (bSuccess)
    {
    __if_exists(WinUSB::CDevice::StopTrackingForTimeSync)
    {
      USB_STOP_TRACKING_FOR_TIME_SYNC_INFORMATION trackingInfo{};
      trackingInfo.TimeTrackingHandle = startTrackingInfo.TimeTrackingHandle;
      bSuccess = device2.StopTrackingForTimeSync(&trackingInfo); //NOLINT(clang-analyzer-deadcode.DeadStores)
    } //__if_exists(WinUSB::CDevice::StopTrackingForTimeSync)
    }
  } //__if_exists(WinUSB::CDevice::StartTrackingForTimeSync)

  //Lets do a simple animation effect with the lights on the LED Bar graph on the FX2 board
  _tprintf(_T("Doing a simple animation effect by turning on and off the lights on the LED bar graph\n"));
  UCHAR bar{0};
  for (ULONG i{0}; i<10; i++)
  {
    for (ULONG j{0}; j<8; j++)
    {
      bar = 1 << j;
      LightBarGraph(device2, bar);
      Sleep(50);
    }
  }

  //Set the LED bar graph to a random set of lights on
  _tprintf(_T("Setting random lights on the LED bar graph to be on / off\n"));
  LightBarGraph(device2, rand() % 256);

  //Determine the state of the switches on the FX2 board
  _tprintf(_T("Checking state of switches\n"));
  USB_DEVICE_REQUESTTYPE requestType;
  requestType.Bits.DataDirection = 1; //Device to host
  requestType.Bits.Recipient = 2; //Endpoint
  requestType.Bits.Type = 2; //Vendor
  WINUSB_SETUP_PACKET packet{};
  packet.RequestType = requestType.bmRequestType; 
  packet.Request = FX2_READ_SWITCHES;
  packet.Index = 0;
  FX2_SWITCH_STATE switchState;
  switchState.SwitchesAsUChar = 0;
  packet.Length = sizeof(switchState);
  packet.Value = 0;
  nBytesReturned = 0;
#pragma warning(suppress: 26490)
  bSuccess = device2.ControlTransfer(packet, reinterpret_cast<PUCHAR>(&switchState), sizeof(switchState), &nBytesReturned, nullptr); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(bSuccess);
#pragma warning(suppress: 26485)
  _tprintf(_T(" Switch8 is %s\n"), switchState.Bits.Switch8 ? _T("ON")  : _T( "OFF"));
#pragma warning(suppress: 26485)
  _tprintf(_T(" Switch7 is %s\n"), switchState.Bits.Switch7 ? _T("ON")  : _T( "OFF"));
#pragma warning(suppress: 26485)
  _tprintf(_T(" Switch6 is %s\n"), switchState.Bits.Switch6 ? _T("ON")  : _T( "OFF"));
#pragma warning(suppress: 26485)
  _tprintf(_T(" Switch5 is %s\n"), switchState.Bits.Switch5 ? _T("ON")  : _T( "OFF"));
#pragma warning(suppress: 26485)
  _tprintf(_T(" Switch4 is %s\n"), switchState.Bits.Switch4 ? _T("ON")  : _T( "OFF"));
#pragma warning(suppress: 26485)
  _tprintf(_T(" Switch3 is %s\n"), switchState.Bits.Switch3 ? _T("ON")  : _T( "OFF"));
#pragma warning(suppress: 26485)
  _tprintf(_T(" Switch2 is %s\n"), switchState.Bits.Switch2 ? _T("ON")  : _T( "OFF"));
#pragma warning(suppress: 26485)
  _tprintf(_T(" Switch1 is %s\n"), switchState.Bits.Switch1 ? _T("ON")  : _T( "OFF"));

  //Determine which lights are on on the LED bar graph on the FX2 board
  _tprintf(_T("Checking which lights are on on the LED bar graph\n"));
  requestType.Bits.DataDirection = 1; //Device to host
  requestType.Bits.Recipient = 2; //Endpoint
  requestType.Bits.Type = 2; //Vendor
  packet.RequestType = requestType.bmRequestType; 
  packet.Request = FX2_READ_BARGRAPH_DISPLAY;
  packet.Index = 0;
  FX2_BAR_GRAPH_STATE barState;
  barState.StateAsUChar = 0;
  packet.Length = sizeof(barState);
  packet.Value = 0;
  nBytesReturned = 0;
#pragma warning(suppress: 26490)
  bSuccess = device2.ControlTransfer(packet, reinterpret_cast<PUCHAR>(&barState), sizeof(barState), &nBytesReturned, nullptr); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(bSuccess);
#pragma warning(suppress: 26485)
  _tprintf(_T(" Bar8 is %s\n"), barState.Bits.Bar8 ? _T("ON")  : _T( "OFF"));
#pragma warning(suppress: 26485)
  _tprintf(_T(" Bar7 is %s\n"), barState.Bits.Bar7 ? _T("ON")  : _T( "OFF"));
#pragma warning(suppress: 26485)
  _tprintf(_T(" Bar6 is %s\n"), barState.Bits.Bar6 ? _T("ON")  : _T( "OFF"));
#pragma warning(suppress: 26485)
  _tprintf(_T(" Bar5 is %s\n"), barState.Bits.Bar5 ? _T("ON")  : _T( "OFF"));
#pragma warning(suppress: 26485)
  _tprintf(_T(" Bar4 is %s\n"), barState.Bits.Bar4 ? _T("ON")  : _T( "OFF"));
#pragma warning(suppress: 26485)
  _tprintf(_T(" Bar3 is %s\n"), barState.Bits.Bar3 ? _T("ON")  : _T( "OFF"));
#pragma warning(suppress: 26485)
  _tprintf(_T(" Bar2 is %s\n"), barState.Bits.Bar2 ? _T("ON")  : _T( "OFF"));
#pragma warning(suppress: 26485)
  _tprintf(_T(" Bar1 is %s\n"), barState.Bits.Bar1 ? _T("ON")  : _T( "OFF"));

  //Determine which lights are on on the 7 Segment display on the FX2 board
  _tprintf(_T("Checking which lights are on on the 7 Segment display\n"));
  requestType.Bits.DataDirection = 1; //Device to host
  requestType.Bits.Recipient = 2; //Endpoint
  requestType.Bits.Type = 2; //Vendor
  packet.RequestType = requestType.bmRequestType; 
  packet.Request = FX2_READ_7SEGMENT_DISPLAY;
  packet.Index = 0;
  UCHAR origSevenSegmentState{0};
  packet.Length = sizeof(origSevenSegmentState);
  packet.Value = 0;
  nBytesReturned = 0;
  bSuccess = device2.ControlTransfer(packet, &origSevenSegmentState, sizeof(origSevenSegmentState), &nBytesReturned, nullptr); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(bSuccess);
  _tprintf(_T(" 7 Segment mask is 0x%X\n"), origSevenSegmentState);

  _tprintf(_T("Cycling lights on the Seven Segment display\n"));
  UCHAR sevenSegment{0};
  for (ULONG i{0}; i<10; i++)
  {
    for (ULONG j{0}; j<8; j++)
    {
      sevenSegment = 1 << j;
      LightSevenSegment(device2, sevenSegment);
      Sleep(50);
    }
  }

  //Reset the 7 Segment display to the original state
  Sleep(4000);
  _tprintf(_T("Resetting lights on the Seven Segment display to original state\n"));
  LightSevenSegment(device2, origSevenSegmentState);

  _tprintf(_T("Please toggle some of the switches on the board\n"));
  for (int i{0}; i<2; i++)
  {
    _tprintf(_T(" Waiting for a switch to be toggled, Loop:%d\n"), i+1);
    switchState.SwitchesAsUChar = 0;
#pragma warning(suppress: 26490)
    bSuccess = device2.ReadPipe(FX2_INTERRUPT_ENDPOINT, reinterpret_cast<PUCHAR>(&switchState), sizeof(switchState), &nBytesReturned, nullptr); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
    ATLASSERT(bSuccess);
#pragma warning(suppress: 26485)
    _tprintf(_T("  Switch8 is %s\n"), switchState.Bits.Switch8 ? _T("ON")  : _T( "OFF"));
#pragma warning(suppress: 26485)
    _tprintf(_T("  Switch7 is %s\n"), switchState.Bits.Switch7 ? _T("ON")  : _T( "OFF"));
#pragma warning(suppress: 26485)
    _tprintf(_T("  Switch6 is %s\n"), switchState.Bits.Switch6 ? _T("ON")  : _T( "OFF"));
#pragma warning(suppress: 26485)
    _tprintf(_T("  Switch5 is %s\n"), switchState.Bits.Switch5 ? _T("ON")  : _T( "OFF"));
#pragma warning(suppress: 26485)
    _tprintf(_T("  Switch4 is %s\n"), switchState.Bits.Switch4 ? _T("ON")  : _T( "OFF"));
#pragma warning(suppress: 26485)
    _tprintf(_T("  Switch3 is %s\n"), switchState.Bits.Switch3 ? _T("ON")  : _T( "OFF"));
#pragma warning(suppress: 26485)
    _tprintf(_T("  Switch2 is %s\n"), switchState.Bits.Switch2 ? _T("ON")  : _T( "OFF"));
#pragma warning(suppress: 26485)
    _tprintf(_T("  Switch1 is %s\n"), switchState.Bits.Switch1 ? _T("ON")  : _T( "OFF"));
  }

  //Reset and flush the bulk in and out endpoints
  bSuccess = device2.AbortPipe(FX2_BULK_IN_ENDPOINT); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(bSuccess);
  bSuccess = device2.ResetPipe(FX2_BULK_IN_ENDPOINT); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(bSuccess);
  bSuccess = device2.FlushPipe(FX2_BULK_IN_ENDPOINT); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(bSuccess);
  bSuccess = device2.ResetPipe(FX2_BULK_OUT_ENDPOINT); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(bSuccess);
  bSuccess = device2.FlushPipe(FX2_BULK_OUT_ENDPOINT); //NOLINT(clang-analyzer-deadcode.DeadStores)

  BOOL bParameter{TRUE};
  bSuccess = device2.SetPowerPolicy(AUTO_SUSPEND, sizeof(bParameter), &bParameter); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(bSuccess);
  ULONG nValueLength{sizeof(bParameter)};
  bParameter = FALSE;
  bSuccess = device2.GetPowerPolicy(AUTO_SUSPEND, &nValueLength, &bParameter); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(bSuccess);

  //Test out bulk IO 
  _tprintf(_T("Testing writing and reading on Bulk endpoints. LED Bar graph should increase as data is being written / received\n"));
  for (int i{0}; i<200; i++)
  { 
    //Create a random output buffer
    std::array<UCHAR, 50> outBuffer{};
    for (auto& b : outBuffer)
      b = rand() % 256;
    ULONG nBytesWritten{0};
#pragma warning(suppress: 26472)
    bSuccess = device2.WritePipe(FX2_BULK_OUT_ENDPOINT, outBuffer.data(), static_cast<ULONG>(outBuffer.size()), &nBytesWritten, nullptr); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
    ATLASSERT(bSuccess);
    std::array<UCHAR, 50> inBuffer{};
    ULONG nBytesRead{0};
#pragma warning(suppress: 26472)
    bSuccess = device2.ReadPipe(FX2_BULK_IN_ENDPOINT, inBuffer.data(), static_cast<ULONG>(inBuffer.size()), &nBytesRead, nullptr); //NOLINT(clang-analyzer-deadcode.DeadStores)
    _tprintf(_T("*"));
    Sleep(50);
#pragma warning(suppress: 26477)
    ATLASSERT(bSuccess);
#pragma warning(suppress: 26477)
    ATLASSERT(nBytesRead == nBytesWritten);
  }
  _tprintf(_T("\nProgram finished. Will now exit\n"));
  Sleep(1000);

  return 0;
}
